package flow.components

import chisel3._
import chisel3.util._
import shapeless.{HNil, ::}
import flow.Params
import flow.components.util._
import chisel3.util.experimental.decode.{decoder, TruthTable}

object ALUControlInterface {
  object OpSelect extends ChiselEnum {
    val aOpAdd, aOpSub, aOpAnd, aOpOr, aOpXor, aOpSlt, aOpSltu, aOpSll, aOpSrl,
        aOpSra = Value
  }
  object SrcASelect extends ChiselEnum {
    val aSrcARs1, aSrcAPc, aSrcAZero = Value
  }
  object SrcBSelect extends ChiselEnum {
    val aSrcBRs2, aSrcBImmI, aSrcBImmJ, aSrcBImmS, aSrcBImmU = Value
  }
  class ALUControlInterface extends Bundle {
    val op = Input(OpSelect())
    val srcASelect = Input(SrcASelect())
    val srcBSelect = Input(SrcBSelect())
    val signExt = Input(Bool())

    def ctrlBindPorts = {
      op :: srcASelect :: srcBSelect :: signExt :: HNil
    }
  }

  def apply() = new ALUControlInterface;
}

class ALU[T <: UInt](tpe: T) extends Module {
  import ALUControlInterface._

  val control = IO(ALUControlInterface())
  val in = IO(new Bundle {
    val a = Input(Vec(SrcASelect.all.length, tpe))
    val b = Input(Vec(SrcBSelect.all.length, tpe))
  })
  val out = IO(new Bundle {
    val eq = Output(Bool())
    val result = Output(tpe)
  })

  val a = in.a(control.srcASelect.asUInt)
  val b = in.b(control.srcBSelect.asUInt)

  // val adder_b = (Fill(tpe.getWidth, io.op(0)) ^ io.b) + io.op(0)  // take (-b) if sub
  val add = a + b
  val sub = a - b
  val and = a & b
  val not = ~a
  val or = a | b
  val xor = a ^ b
  val slt = a.asSInt < b.asSInt
  val sltu = a < b
  val sll = a << b(log2Ceil(tpe.getWidth), 0)
  val srl = a >> b(log2Ceil(tpe.getWidth), 0)
  val sra = a.asSInt >> b(log2Ceil(tpe.getWidth), 0)
  out.eq := a === b

  import ALUControlInterface.OpSelect._

  out.result := MuxLookup(control.op, 0.U)(
    Seq(
      aOpAdd -> add,
      aOpSub -> sub,
      aOpAnd -> and,
      aOpOr -> or,
      aOpXor -> xor,
      aOpSlt -> slt,
      aOpSltu -> sltu,
      aOpSll -> sll,
      aOpSrl -> srl,
      aOpSra -> sra.asUInt
    )
  )
}

object ALU {
  def apply[T <: UInt](tpe: T): ALU[T] = {
    Module(new ALU(tpe))
  }
}

class newALU(implicit p: Params) extends Module {
  import ALUControlInterface._
  val control = IO(ALUControlInterface())
  val in = IO(new Bundle {
    val a = Input(Vec(SrcASelect.all.length, UInt(p.XLEN)))
    val b = Input(Vec(SrcBSelect.all.length, UInt(p.XLEN)))
  })
  val out = IO(new Bundle {
    val isEqual = Output(Bool())
    val result = Output(UInt(p.XLEN))
  })

  val a = in.a(control.srcASelect.asUInt)
  val b = in.b(control.srcBSelect.asUInt)

  // val adder_b = (Fill(tpe.getWidth, io.op(0)) ^ io.b) + io.op(0)  // take (-b) if sub
  val add = a + b
  val sub = a - b
  val and = a & b
  val not = ~a
  val or = a | b
  val xor = a ^ b
  val slt = a.asSInt < b.asSInt
  val sltu = a < b
  val sll = a << b(5, 0)
  val srl = a >> b(5, 0)
  val sra = a.asSInt >> b(5, 0)
  out.isEqual := a === b

  import OpSelect._

  out.result := MuxLookup(control.op, 0.U)(
    Seq(
      aOpAdd -> add,
      aOpSub -> sub,
      aOpAnd -> and,
      aOpOr -> or,
      aOpXor -> xor,
      aOpSlt -> slt,
      aOpSltu -> sltu,
      aOpSll -> sll,
      aOpSrl -> srl,
      aOpSra -> sra.asUInt
    )
  )
}

class AluController(implicit p: Params) extends Module {
  val out = IO(Flipped(ALUControlInterface()))
  val in = IO(new Bundle {
    val inst = Input(UInt(p.XLEN))
  })
  import flow.components.RV32Inst._
  import ALUControlInterface._
  // OpSelect
  // format: off
  val aOpAdd = Array(
    lui, auipc,
    jal, jalr,
    lb, lbu, lh, lhu, lw,
    sb, sh, sw,
    addi, add
  ).map(_ -> OpSelect.aOpAdd.BP)
  // format: on

  val aOpAnd = Array(and, andi).map(_ -> OpSelect.aOpAnd.BP)

  val aOpSub = Array(sub).map(_ -> OpSelect.aOpSub.BP)

  val aOpOr = Array(ori, or).map(_ -> OpSelect.aOpOr.BP)

  val aOpXor = Array(xori, xor).map(_ -> OpSelect.aOpXor.BP)

  val aOpSlt = Array(beq, bne, blt, bge, slti, slt).map(_ -> OpSelect.aOpSlt.BP)

  val aOpSltu = Array(sltiu, sltu, bltu, bgeu).map(_ -> OpSelect.aOpSltu.BP)

  val aOpSll = Array(slli, sll).map(_ -> OpSelect.aOpSll.BP)

  val aOpSrl = Array(srli, srl).map(_ -> OpSelect.aOpSrl.BP)

  val aOpSra = Array(
    sra,
    srai
  ).map(_ -> OpSelect.aOpSra.BP)

  val OpCodeMapping = TruthTable(
    aOpAdd ++ aOpAnd ++ aOpSub ++ aOpOr ++ aOpXor ++ aOpSlt ++ aOpSltu ++ aOpSll ++ aOpSrl ++ aOpSra,
    BitPat.dontCare(OpSelect.getWidth)
  )

  val aSrcAZero = Array(
    lui
  ).map(_ -> SrcASelect.aSrcAZero.BP)

  val aSrcAPc = Array(
    auipc,
    jal
  ).map(_ -> SrcASelect.aSrcAPc.BP)

  val srcAMapping = TruthTable(
    aSrcAZero ++ aSrcAPc,
    SrcASelect.aSrcARs1.BP
  )
  val aSrcBImmS = Array(sb, sh, sw).map(_ -> SrcBSelect.aSrcBImmS.BP)

  val aSrcBImmU = Array(lui, auipc).map(_ -> SrcBSelect.aSrcBImmU.BP)

  val aSrcBImmJ = Array(jal).map(_ -> SrcBSelect.aSrcBImmJ.BP)

  // format: off
  val aSrcBImmI = Array(
    jalr,
    lb, lbu, lh, lhu, lw,
    addi, slti, sltiu, xori, ori, andi, slli, srli, srai
  ).map(_ -> SrcBSelect.aSrcBImmI.BP)
  // format: on

  val srcBMapping = TruthTable(
    aSrcBImmI ++ aSrcBImmJ ++ aSrcBImmU ++ aSrcBImmS,
    SrcBSelect.aSrcBRs2.BP
  )

  val doSignExt = Array(blt, bge, slt, slti).map(_ -> BitPat(true.B))

  val noSignExt = Array(lui, auipc, jal, jalr, bltu, bgeu, sltiu, sltu).map(
    _ -> BitPat(false.B)
  )

  val signExtMapping = TruthTable(doSignExt ++ noSignExt, BitPat.dontCare(1))

  out.op := OpSelect.safe(decoder(in.inst, OpCodeMapping))._1
  out.srcASelect := SrcASelect.safe(decoder(in.inst, srcAMapping))._1
  out.srcBSelect := SrcBSelect.safe(decoder(in.inst, srcBMapping))._1
  out.signExt := decoder(in.inst, signExtMapping)
}
