package flow.components

import chisel3._
import chisel3.util.log2Ceil
import chisel3.util.UIntToOH
import chisel3.util.MuxLookup
import chisel3.util.BitPat
import chisel3.util.experimental.decode.{decoder, TruthTable}
import chisel3.experimental.Trace._
import shapeless.{HList, HNil, ::}
import flow.Params
import flow.components.util._
import flow.components.RV32Inst._

object RegControl {
  object WriteSelect extends ChiselEnum {
    val rAluOut, rMemOut, rNpc = Value
  }
  class RegControl extends Bundle {
    val writeEnable = Input(Bool())
    val writeSelect = Input(WriteSelect())

    def ctrlBindPorts = {
      writeEnable :: writeSelect :: HNil
    }
  }

  def apply() = new RegControl
}

class RegisterFile[T <: Data](tpe: T, regCount: Int, numReadPorts: Int)
    extends Module {
  require(numReadPorts >= 0)
  val control = IO(RegControl())
  val dataAddrWidth = log2Ceil(regCount).W
  val in = IO(new Bundle {
    val writeAddr = Input(UInt(dataAddrWidth))
    val writeData = Input(Vec(RegControl.WriteSelect.all.length, tpe))
    val rs = Input(Vec(numReadPorts, UInt(dataAddrWidth)))
  })
  val out = IO(new Bundle {
    val src = Output(Vec(numReadPorts, tpe))
  })

  val regResetValue = 0.U(tpe.getWidth.W)
  val regFile = RegInit(VecInit(Seq.fill(regCount)(regResetValue)))
  val writeAddrOH = UIntToOH(in.writeAddr)

  for ((reg, i) <- regFile.zipWithIndex.tail) {
    reg := Mux(
      writeAddrOH(i.U(log2Ceil(regCount).W)) && control.writeEnable,
      in.writeData(control.writeSelect.asUInt),
      reg
    )
  }
  regFile(0) := 0.U

  for (port <- 0 until numReadPorts) {
    out.src(port) := regFile(in.rs(port).asUInt)
  }
}

class newRegisterFile(implicit p: Params) extends Module {
  val in = IO(new Bundle {
    val rd = Input(UInt(p.regsAddrWidth))
    val writeData = Input(Vec(RegControl.WriteSelect.all.length, UInt(p.XLEN)))
    val rs1 = Input(UInt(p.regsAddrWidth))
    val rs2 = Input(UInt(p.regsAddrWidth))
  })

  val out = IO(new Bundle {
    val src1 = Output(UInt(p.XLEN))
    val src2 = Output(UInt(p.XLEN))
  })

  val control = IO(RegControl())

  val regFile = RegInit(
    VecInit(Seq.fill(p.regsCount)(p.regsResetValue.U(p.XLEN)))
  )

  // 0.U -> writeData(0), ...
  val writeDataMux =
    MuxBindCtrlSrc(control.writeSelect, in.writeData, regFile(in.rd))

  regFile(in.rd) := Mux(
    control.writeEnable,
    writeDataMux,
    regFile(in.rd)
  )
  regFile(0) := 0.U

  out.src1 := regFile(in.rs1)
  out.src2 := regFile(in.rs2)
  traceName(regFile)
  dontTouch(regFile)
}

class RegisterFileController(implicit p: Params) extends Module {
  val in = IO(new Bundle {
    val inst = Input(UInt(p.XLEN))
  })

  val out = IO(Flipped(RegControl()))

  // format: off
  import RegControl._
  private val aluOutMapping =
    Array(
      lui, auipc,
      addi, slti, sltiu, xori, ori, andi, slli, srli, srai,
      add, sub, sll, slt, sltu, xor, srl, sra, or, and,
    ).map(_ -> WriteSelect.rAluOut.BP)

  private val memOutMapping =
    Array(
      lb, lbu, lh, lhu, lw
    ).map(_ -> WriteSelect.rMemOut.BP)

  private val npcMapping =
    Array(
      jal, jalr
    ).map(_ -> WriteSelect.rNpc.BP)
  // format: on

  private val writeSelectMapping = TruthTable(
    aluOutMapping ++ memOutMapping ++ npcMapping,
    BitPat.dontCare(out.writeSelect.getWidth)
  )

  // enable write if instruction belongs to any mapping above
  private val writeEnableMapping = TruthTable(
    (aluOutMapping ++ memOutMapping ++ npcMapping).map(x =>
      (x._1 -> BitPat("b1"))
    ),
    BitPat("b0")
  )

  println(writeEnableMapping)

  out.writeSelect := RegControl.WriteSelect
    .safe(
      decoder(in.inst, writeSelectMapping)
    )
    ._1
  out.writeEnable := decoder(in.inst, writeEnableMapping)
}
